/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include "gauntlet.h"

void read_map(char *filename) {
  FILE *fp;
  int i,j;
  char buf[MAX_BUF];
  object *op;

  if((fp=fopen(filename,"r"))==NULL) {
    (void) fprintf(stderr,"Can't open %s.\n",filename);
    exit(1);
  }
  for(i=0;i<MAX_X;i++)
    for(j=0;j<MAX_Y;j++)
      omap[i][j]=G_BLANK,map[i][j]=G_BLANK,map_ob[i][j]=NULL;
  j=0,mapx=0,mapy=80,startx=2,starty=2;
  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    if(strlen(buf)-2>mapx) mapx=strlen(buf)-2;
    for(i=0;i<MAX_X&&buf[i]!='\0'&&buf[i]!='\n';i++) {
      switch(buf[i]) {
      case G_START:
        startx=i,starty=j;
        omap[i][j]=G_BLANK,map[i][j]=G_BLANK;
        break;
      case G_TREASURE:
        op=get_object();
        op->type=TREASURE;
        op->value=7;
        op->x=i,op->y=j;
        op->face=G_TREASURE;
        op->speed=0;
        op->speed_left=0;
        op->weight=30;
        strcpy(op->name,"chest");
        omap[i][j]=G_BLANK;
        insert_ob_in_map(op);
        break;
      case G_POTION:
        op=get_object();
        op->type=POTION;
        op->x=i,op->y=j;
        op->face=G_POTION;
        op->speed=0,op->speed_left=0;
        op->weight=2;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"potion");
        insert_ob_in_map(op);
        break;
      case G_FOOD:
        op=get_object();
        if(random()%2) {
          op->type=FOOD;
          op->stats.food=200;
          op->value=10;
          op->face=G_FOOD;
        } else {
          op->type=POISON;
          op->stats.hp=-3;
          op->face=G_POISON;
          op->value=1;
        }
        op->x=i,op->y=j;
        op->speed=0,op->speed_left=0;
        op->weight=18;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"food");
        insert_ob_in_map(op);
        break;
      case G_KEY:
        op=get_object();
        op->type=KEY;
        op->value=50;
        op->x=i,op->y=j;
        op->face=G_KEY;
        op->speed=0;
        op->weight=1;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"key");
        insert_ob_in_map(op);
        break;
      case G_GENGOB:
        op=get_object();
        op->type=GENGOB;
        op->speed=0.02,op->speed_left=op->speed;
        op->face=G_GENGOB;
        op->stats.hp=10;
        op->stats.ac=8;
        op->stats.exp=100;
        op->x=i,op->y=j;
        op->alive=1;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"generator");
        insert_ob_in_map(op);
        break;
      case G_GOB1:
        op=get_object();
        op->alive=1;
        op->type=GOB;
        op->speed=0.1,op->speed_left=op->speed;
        op->x=i,op->y=j;
        op->face=G_GOB1;
        op->stats.hp=3;
        op->stats.wc=18;
        op->stats.dam=3;
        op->stats.ac=10;
        op->stats.exp=5;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"goblin");
        insert_ob_in_map(op);
        break;
      case G_DOOR_1:
        op=get_object();
        op->alive=1;
        op->type=DOOR;
        op->speed=0,op->speed_left=0;
        op->x=i,op->y=j;
        op->face=G_DOOR_1;
        op->stats.hp=300;
        op->stats.ac=10;
        op->stats.exp=1;
        omap[i][j]=G_KEY;
        strcpy(op->name,"door");
        insert_ob_in_map(op);
        break;
      case G_DOOR_2:
        op=get_object();
        op->alive=1;
        op->type=DOOR;
        op->speed=0,op->speed_left=0;
        op->x=i,op->y=j;
        op->face=G_DOOR_2;
        op->stats.hp=300;
        op->stats.ac=10;
        op->stats.exp=1;
        omap[i][j]=G_KEY;
        strcpy(op->name,"door");
        insert_ob_in_map(op);
        break;
      case G_GENMUNCH:
        op=get_object();
        op->alive=1;
        op->type=GENMUNCH;
        op->speed=0.002,op->speed_left=0;
        op->face=G_GENMUNCH;
        op->x=i,op->y=j;
        op->stats.hp=30;
        op->stats.ac=7;
        op->stats.exp=300;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"genmunch");
        insert_ob_in_map(op);
        break;
      case G_NEST:
        op=get_object();
        op->alive=1;
        op->type=NEST;
        op->speed=0.01,op->speed_left=0;
        op->face=G_NEST;
        op->x=i,op->y=j;
        op->stats.hp=20;
        op->stats.ac=5;
        op->stats.exp=200;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"nest");
        insert_ob_in_map(op);
        break;
      case G_GENSCORP:
        op=get_object();
        op->alive=1;
        op->type=GENSCORP;
        op->speed=0.002,op->speed_left=0;
        op->face=G_GENSCORP;
        op->x=i,op->y=j;
        op->stats.hp=4;
        op->stats.ac=3;
        op->stats.exp=50;
        omap[i][j]=G_BLANK;
        strcpy(op->name,"genscorp");
        insert_ob_in_map(op);
        break;
      default:
        omap[i][j]=buf[i],map[i][j]=buf[i];
        break;
      }
    }
    if(i>mapx) mapx=--i;
    j++;
  }
  mapy=--j;
}
