/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include "version.h"
#include "gauntlet.h"
#include "main.h"

extern char *get_root_display();
extern char *get_stats_display();
extern char *get_info_display();
extern char *get_game_display();
extern int IOerrors();
extern int Xerrors();
extern char ann_last[MAX_BUF];


extern object *objects;
extern object *free_objects;
extern object objarray[STARTMAX];

void dump_map() {
  int i,j;
  for(i=0;i<=mapy;i++) {
    if(map[0][i]=='\0')
      break;
    for(j=0;j<=mapx;j++) {
      if(map[j][i]=='\0')
        break;
      (void) fprintf(stderr,"%c",map[j][i]);
    }
    (void) fprintf(stderr,"\n");
  }
}

void dump_omap() {
  int i,j;
  for(i=0;i<=mapy;i++) {
    for(j=0;j<=mapx;j++)
      (void) fprintf(stderr,"%c",omap[j][i]);
    (void) fprintf(stderr,"\n");
  }
}

void start_info(object *op) {
  char buf[MAX_BUF];

  draw_info(op,"Welcome to Gauntlet, v0.1!");
  draw_info(op,"Push `hjklynub' to walk in a direction.");
  draw_info(op,"Shift + direction = fire, Ctrl + direction = run.");
  draw_info(op,"You damage monsters which you try to walk into.");
  draw_info(op,"`i' = inventory, `:' = look, `,' = pick, `d' = drop.");
  draw_info(op,"`>' = rotate inventory, 'a' = apply top of inventory.");
  draw_info(op,"'+' = change spell.");
  draw_info(op,"Type \"'help\" for help about extended commands.");
  draw_info(op,"Type `\"' to speak.");
  draw_info(op," ");
  (void) sprintf(buf,"%s entered the game.",op->name);
  info_all(buf);
  info_flush();
}

int add_player(char *disp) {
  player *p;
  char buf[MAX_BUF];
  int i;

  i=find_free_player();
  if(i==-1) {
    (void) sprintf(errmsg,"There is not room for more players.");
    return 1;
  }
  p=&players[i];
  strncpy(p->name,disp,MAX_NAME);
  p->writing=0;
  strncpy(buf,disp,MAX_BUF);
  if(strchr(buf,':')==NULL)
    strcat(buf,":0");
  if(get_root_display(p,buf))
    return 1;
  if(get_game_display(p,buf))
    return 1;
  if(get_stats_display(p,buf))
    return 1;
  if(get_info_display(p,buf))
    return 1;
  p->ob=get_player(p);
  draw_info(p->ob,"Welcome, Brave Warrior!");
  draw_info(p->ob," ");
  draw_info(p->ob,"Roll again (y/n)?");
  strncpy(p->ob->name,disp,MAX_NAME);
  p->used=1;
  return 0;
}

void init() {
  int i,k;

  strcpy(ann_last,"blah");
/* Initialize all objects: */
  free_objects=objarray;
  objects=NULL;
  objarray[0].prev=NULL,
  objarray[0].next=&objarray[1],
  objarray[0].is_freed=1,
  objarray[0].is_removed=1;
  for(i=1;i<STARTMAX-1;i++)
    objarray[i].next=&objarray[i+1],
    objarray[i].prev=&objarray[i-1],
    objarray[i].is_freed=1,
    objarray[i].is_removed=1;
  objarray[STARTMAX-1].next=NULL;
  objarray[STARTMAX-1].prev=&objarray[STARTMAX-2];
  objarray[STARTMAX-1].is_freed=1;
  objarray[STARTMAX-1].is_removed=1;
  

  for(i=0;i<MAXPLAYERS;i++) {
    players[i].used=0,players[i].nr=i,players[i].infoline=0;
    for(k=0;k<INFOLINES;k++)
      strncpy(players[i].info," ",30);
  }
  read_map("gmap");

  if(gargc<=1||gargc>=MAXPLAYERS) {
    (void) fprintf(stderr,"Usage: gauntlet [display1] [display2]...\n");
    exit(1);
  }
  XSetIOErrorHandler(IOerrors);
  XSetErrorHandler(Xerrors);
  for(i=1;i<gargc;i++)
    if(add_player(gargv[i]))
      (void) fprintf(stderr,"%s\n",errmsg);
  if(!players[0].used)
    exit(1);
}

int find_free_player() {
  int i;

  for(i=0;i<MAXPLAYERS;i++)
    if(!players[i].used)
      return i;
  return -1;
}

int find_free_spot(int x,int y,int start,int stop) {
  int i;
  for(i=start;i<stop;i++)
    if(!blocked(x+freearr_x[i],y+freearr_y[i]))
      return i;
    else if(wall(x+freearr_x[i],y+freearr_y[i])&&maxfree[i]<stop)
      stop=maxfree[i];
  return 0;
}

int find_dir(int x, int y, object *exclude) {
  int i,max=SIZEOFFREE;
  object *tmp;

  for(i=1;i<max;i++) {
    if(wall(x+freearr_x[i],y+freearr_y[i]))
    ;
    else {
      tmp=map_ob[x+freearr_x[i]][y+freearr_y[i]];
      while(tmp!=NULL&&!tmp->alive)
        tmp=tmp->above;
      if(tmp&&tmp->alive&&tmp!=exclude&&tmp->type!=DOOR) {
        return freedir[i];
      }
    }
  }
  return 0;
}

int roll_stat() {
  int a[4],i,j,k;
  for(i=0;i<4;i++)
    a[i]=(int)random()%6+1;
  for(i=0,j=0,k=7;i<4;i++) 
    if(a[i]<k)
      k=a[i],j=i;
  for(i=0,k=0;i<4;i++) {
    if(i!=j)
      k+=a[i];
  }
  return k<84?k:roll_stat();
}

void roll_stats(object *op) {
  op->stats.Str=roll_stat();
  op->stats.Dex=roll_stat();
  op->stats.Int=roll_stat();
  op->stats.Con=roll_stat();
  op->stats.Wis=roll_stat();
  op->stats.Cha=roll_stat();
  op->stats.hp=0;
  op->stats.level=0;
  op->stats.exp=0;
  op->stats.sp=0;
  op->stats.ac=0;
  op->stats.level=0;
  add_exp(op,0);
  op->stats.sp=op->stats.maxsp;
  op->stats.hp=op->stats.maxhp;
}

object *get_player(player *p) {
  object *op=get_object();
  int i,j;

  i=find_free_spot(startx,starty,0,SIZEOFFREE);
  op->x=startx+freearr_x[i],op->y=starty+freearr_y[i];
  op->face=G_CHAR;
  op->type=PLAYER;
  op->alive=1;
  op->speed=0.5;
  op->speed_left=0.5;
  op->dirx=0,op->diry=0;
  op->contr=p;
  op->stats.wc=2;
  op->stats.exp=0;
  op->stats.food=500;
  roll_stats(op);
  op->state=2;
  p->last_stats.Str=0, p->last_stats.Dex=0,
  p->last_stats.Int=0, p->last_stats.Con=0,
  p->last_stats.Wis=0, p->last_stats.Cha=0;
  p->last_stats.hp=0,  p->last_stats.maxhp=0;
  p->last_stats.wc=0,  p->last_stats.ac=0;
  p->last_stats.sp=0,  p->last_stats.maxsp=0;
  p->last_stats.exp=-1,p->last_stats.food=0;
  p->last_value=-1;
  p->last_speed=-1;
  p->shoottype=0,p->shootstrength=5;
  for(i=WINLEFT;i<=WINRIGHT;i++)
    for(j=WINUPPER;j<=WINLOWER;j++)
      op->contr->drawn[i-WINLEFT][j-WINUPPER]='{'; /* No such character used */
  return op;
}

object *create_monster() {
  object *op=get_object();
  op->x=startx,op->y=starty;
  op->face=G_GOB1;
  op->type=GOB;
  op->speed=0.1;
  op->speed_left=-0.001;
  op->alive=1;
  op->stats.wc=1;
  op->stats.ac=1;
  op->stats.hp=1;
  op->stats.exp=20;
  strcpy(op->name,"goblin");
  return op;
}

int distance(object *ob1,object *ob2) {
  int i;
  i= (ob1->x - ob2->x)*(ob1->x - ob2->x)+
         (ob1->y - ob2->y)*(ob1->y - ob2->y);
  return i;
}

object *get_nearest_player(object *ob) {
  object *pl;
  int i,lastdist,tmp;

  pl=NULL;
  for(i=0,lastdist=1000;i<MAXPLAYERS;i++) {
    if(!players[i].used||players[i].ob->state)
      continue;
    tmp=distance(players[i].ob,ob);
    if(lastdist>tmp) {
      pl=players[i].ob;
      lastdist=tmp;
    }
  }
  return pl;
}

void remove_door(object *op) {
  int i;
  object *tmp;
  for(i=1;i<9;i+=2)
    if((tmp=present(DOOR,op->x+freearr_x[i],op->y+freearr_y[i]))!=NULL)
      tmp->speed=0.1,tmp->speed_left=0;
  omap[op->x][op->y]=G_BLANK;
  remove_ob(op);
  free_object(op);
}
   

void generate_gob(object *gen) {
  int i;
  object *op;

  i=find_free_spot(gen->x,gen->y,1,9);
  if(!i) return;
  op=create_monster();
  op->stats.ac=15;
  op->stats.wc=17;
  op->stats.dam=4;
  op->stats.hp=3;
  op->type=GOB;
  op->x=gen->x+freearr_x[i];
  op->y=gen->y+freearr_y[i];
  strcpy(op->name,"goblin");
  insert_ob_in_map(op);
}

void generate_munch(object *gen) {
  int i;
  object *op;

  i=find_free_spot(gen->x,gen->y,1,9);
  if(!i) return;
  op=create_monster();
  op->stats.exp=100;
  op->stats.ac=10;
  op->stats.hp=60;
  op->stats.wc=2;
  op->stats.dam=10;
  op->speed=0.05;
  op->type=MUNCH;
  op->x=gen->x+freearr_x[i];
  op->y=gen->y+freearr_y[i];
  strcpy(op->name,"muncher");
  insert_ob_in_map(op);
}

void generate_bird(object *gen) {
  int i;
  object *op;

  i=find_free_spot(gen->x,gen->y,1,9);
  if(!i) return;
  op=create_monster();
  op->stats.ac=2;
  op->stats.hp=2;
  op->stats.wc=14;
  op->stats.dam=1;
  op->speed=0.3;
  op->stats.exp=20;
  op->type=BIRD;
  op->x=gen->x+freearr_x[i];
  op->y=gen->y+freearr_y[i];
  strcpy(op->name,"bird");
  insert_ob_in_map(op);
}

void generate_scorp(object *gen) {
  int i;
  object *op;

  i=find_free_spot(gen->x,gen->y,1,9);
  if(!i) return;
  op=create_monster();
  op->stats.ac=10;
  op->stats.hp=10;
  op->stats.wc=14;
  op->stats.dam=999;
  op->speed=0.04;
  op->stats.exp=70;
  op->type=SCORP;
  op->x=gen->x+freearr_x[i];
  op->y=gen->y+freearr_y[i];
  strcpy(op->name,"scorpion");
  insert_ob_in_map(op);
}

void attack_ob(object *op,object *hitter) {
  int roll;
  char buf[MAX_BUF];
  roll=random()%20+1;
  if(roll==20||op->stats.ac>=hitter->stats.wc-roll) {
    if(op->type==PLAYER&&(hitter->owner==NULL?hitter->type:hitter->owner->type)
       ==PLAYER) {
      if(hitter->owner!=NULL)
        sprintf(buf,"%s hits you with %s",hitter->owner->name,hitter->name);
      else
        sprintf(buf,"%s hits you.",hitter->name);
      draw_info(op,buf);
    }
    if(hitter->type==PLAYER) {
      sprintf(buf,"You hit %s.",op->name);
      draw_info(hitter,buf);
    } else
    if(hitter->owner!=NULL&&hitter->owner->type==PLAYER) {
      sprintf(buf,"You hit %s with %s.",op->name,hitter->name);
      draw_info(hitter->owner,buf);
    }
    hit_player(op,random()%(hitter->stats.dam+1)+1,hitter);
  } else {
    if(op->type==PLAYER&&(hitter->owner==NULL?hitter->type:hitter->owner->type)
       ==PLAYER) {
       if(hitter->owner!=NULL)
         sprintf(buf,"%s missed you with %s",hitter->owner->name,hitter->name);
       else
         sprintf(buf,"%s missed you.",hitter->name);
       draw_info(op,buf);
    }
    if(hitter->type==PLAYER) {
      sprintf(buf,"You missed %s.",op->name);
      draw_info(hitter,buf);
    } else
    if(hitter->owner!=NULL&&hitter->owner->type==PLAYER) {
      sprintf(buf,"You missed %s with your %s.",op->name,hitter->name);
      draw_info(hitter->owner,buf);
    }
  }
}

int move_monster(object *op) {
  object *pl;
  int i;

  if(op->stats.hp<0) {
    fprintf(stderr,"Moving dead monster:\n");
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
    return 0;
  }
  switch(op->type) {
  case BIRD:
    op->face=op->state?G_BIRDIE1:G_BIRDIE2;
    break;
  case SCORP:
    op->face=op->state?G_SCORPIE1:G_SCORPIE2;
    break;
  case MUNCH:
    op->face=op->state?G_MUNCH1:G_MUNCH2;
    break;
  case GOB:
    op->face=op->state?G_GOB2:G_GOB1;
    break;
  }
  op->state=op->state?0:1;
  update_object(op);
  pl=get_nearest_player(op);
  if(pl==NULL) return 1;
  if((pl->x>op->x&&move_ob(op, 1, 0))||
     (pl->x<op->x&&move_ob(op,-1, 0))||
     (pl->y<op->y&&move_ob(op, 0,-1))||
     (pl->y>op->y&&move_ob(op, 0, 1))) return 1;
  /* Monster can't move...now see if it can hit the player */
  for(i=1;i<9;i++)
    if(op->x+freearr_x[i]==pl->x && op->y+freearr_y[i]==pl->y) {
      attack_ob(pl,op);
      break;
    }
  return 0;
}

void move_bullet(object *op) {
  remove_ob(op);
  op->x+=op->dirx,op->y+=op->diry;
  if(wall(op->x,op->y)) {
    free_object(op);
    return;
  }
  if(blocked(op->x,op->y)) {
    hit_map(op,0,0);
    free_object(op);
    return;
  }
  insert_ob_in_map(op);
}

int rot_diff(int i, int j) {
  int k;
  k=j-i;
  if(k<-4) k+=8;
  if(k>4) k-=8;
  return k;
}

void move_missile(object *op) {
  int i;

  remove_ob(op);
  op->x+=op->dirx,op->y+=op->diry;
  if(wall(op->x,op->y)) {
    free_object(op);
    return;
  }
  if(blocked(op->x,op->y)) {
    hit_map(op,0,0);
    free_object(op);
    return;
  }
  i=find_dir(op->x,op->y,op->owner);
  if(i&&i!=op->value) {
    op->value+=(rot_diff(op->value,i)>0)?1:-1;
    if(op->value>8) op->value-=8;
    if(op->value<1) op->value+=8;
    op->face=mmface[op->value];
    op->dirx=freearr_x[op->value];
    op->diry=freearr_y[op->value];
  }
  insert_ob_in_map(op);
}

void stop_arrow(object *op,object *tmp) {
  if(wall(op->x,op->y))
    op->x-=op->dirx,op->y-=op->diry;
  op->dirx=0,op->diry=0;
  op->speed=0,op->speed_left=-2;
  insert_ob_in_map(op);
  if(tmp!=NULL&&tmp->stats.hp>=0) { /* Hack to get the monster on top of the arrow */
    remove_ob(tmp);
    insert_ob_in_map(tmp);
  }
}

void move_arrow(object *op) {
  object *tmp;

  remove_ob(op);
  op->x+=op->dirx,op->y+=op->diry;
  if(op->x<0||op->x>mapx||op->y<0||op->y>mapy) {
    fprintf(stderr,"Arrow out of map.\n");
    stop_arrow(op,NULL);
    remove_ob(op);
    free_object(op);
    return;
  }
  if(wall(op->x,op->y)) {
    stop_arrow(op,NULL);
    return;
  }
  tmp=map_ob[op->x][op->y];
  while(tmp!=NULL&&!tmp->alive)
    tmp=tmp->above;
  if(tmp==NULL) {
    insert_ob_in_map(op);
    return;
  }
  attack_ob(tmp,op);
  stop_arrow(op,tmp);
}

void fire(object *op,int dir) {
  object *tmp=get_object();
  int dx=freearr_x[dir],dy=freearr_y[dir];

  tmp->owner=op;
  tmp->dirx=dx,tmp->diry=dy;
  tmp->x=op->x,tmp->y=op->y;
  tmp->speed=1;
  tmp->speed_left=0;
  tmp->stats.hp=9;
  tmp->alive=1;
  switch(op->contr->shoottype) {
  case 0:
    if(!op->wizard&&op->stats.sp<1) {
      free_object(tmp);
      draw_info(op,"You don't have enough spellpoints.");
      return;
    }
    op->stats.sp--;
    tmp->face=G_BULLET;
    tmp->stats.hp=0;
    tmp->type=BULLET;
    tmp->stats.wc=2;
    tmp->stats.dam=4;
    strcpy(tmp->name,"bullet");
    insert_ob_in_map(tmp);
    move_bullet(tmp);
    return;
  case 1:
    if(!op->wizard&&op->stats.sp<op->contr->shootstrength) {
      free_object(tmp);
      draw_info(op,"You don't have enough spellpoints.");
      return;
    }
    op->stats.sp-=op->contr->shootstrength;
    draw_stats(op);
    tmp->stats.hp=999;
    tmp->stats.wc=(op->contr->shootstrength-5)/3+5;
    strcpy(tmp->name,"firebullet");
    tmp->type=FBULLET;
    tmp->face=fbface[dir];
    insert_ob_in_map(tmp);
    move_fb_bullet(tmp);
    return;
  case 2:
    if(!op->wizard&&op->stats.sp<op->contr->shootstrength) {
      free_object(tmp);
      draw_info(op,"You don't have enough spellpoints.");
      return;
    }
    op->stats.sp-=op->contr->shootstrength;
    draw_stats(op);
    tmp->type=LIGHTNING;
    tmp->face=lface[dir];
    tmp->value=dir;
    tmp->state=0;
    tmp->stats.wc=1+(op->contr->shootstrength-5)/10;
    tmp->stats.exp=(op->contr->shootstrength-5)/2+8;
    tmp->stats.hp=4+(op->contr->shootstrength-5)/2;
    strncpy(tmp->name,"lightning",MAX_NAME);
    tmp->alive=0;
    tmp->stats.dam=2;
    tmp->x+=tmp->dirx,tmp->y+=tmp->diry;
    if(wall(tmp->x,tmp->y)) {
      tmp->x=op->x,tmp->y=op->y;
      if((tmp->value+=4)>8)
        tmp->value-=8;
      tmp->dirx=-tmp->dirx,tmp->diry=-tmp->diry;
    }
    fprintf(stderr,"First Lightning:\n");
    insert_ob_in_map(tmp);
    fprintf(stderr,"Done:\n");
    move_lightning(tmp);
    fprintf(stderr,"Moved it.\n");
    return;
  case 3:
    if(!op->wizard&&op->stats.sp<2) {
      free_object(tmp);
      draw_info(op,"You don't have enough spellpoints.");
      return;
    }
    op->stats.sp-=2;
    draw_stats(op);
    tmp->face=mmface[dir];
    tmp->value=dir;
    tmp->stats.hp=0;
    tmp->type=MMISSILE;
    tmp->stats.wc=2;
    tmp->stats.dam=4;
    strcpy(tmp->name,"missile");
    insert_ob_in_map(tmp);
    move_missile(tmp);
    return;
  case 4:
    if(!split_ob(op,ARROW)) {
      free_object(tmp);
      draw_info(op,"You have no arrows left.");
      return;
    }
    fix_player(op);
    tmp->type=ARROW;
    tmp->face=aface[dir];
    strncpy(tmp->name,"arrow",MAX_NAME);
    tmp->stats.dam=random()%(dam_bonus[op->stats.Str]+2)+1;
    tmp->stats.wc=op->stats.wc;
    tmp->alive=0;
    tmp->weight=1;
    tmp->nrof=1;
    insert_ob_in_map(tmp);
    move_arrow(tmp);
    return;
  default:
    free_object(tmp);
    draw_info(op,"Illegal shoot type.");
    return;
  }
}

object *get_lightning(object *l) {
  object *op=get_object();
  if(!op->is_removed) {
    fprintf(stderr,"Lightning: get_object() returned busy object.\n");
    remove_ob(op);
  }
  op->face=l->face;
  op->owner=l->owner;
  op->dirx=l->dirx,op->diry=l->diry;
  op->x=l->x,op->y=l->y;
  op->speed=l->speed,op->speed_left=-1.1;
  op->stats.hp=l->stats.hp+1;
  op->alive=0;
  op->value=l->value;
  op->stats.exp=l->stats.exp;
  op->stats.wc=l->stats.wc;
  op->stats.dam=l->stats.dam;
  op->state=0;
  op->type=LIGHTNING;
  strncpy(op->name,l->name,MAX_NAME);
  return op;
}

void move_lightning(object *l) {
  object *op;
  if(--(l->stats.hp)<0) {
    remove_ob(l);
    free_object(l);
    return;
  }
  if(!l->state&&--(l->stats.exp)>0) {
    l->state=1;
    if(wall(l->x+l->dirx,l->y+l->diry)) {
      l->dirx=-l->dirx,l->diry=-l->diry,l->state=0;
      l->face+=4;
      if(l->face>G_LIGHT_8)
        l->face-=8;
      update_object(l);
      return;
    }
    else {
      op=get_lightning(l);
      op->x+=l->dirx,op->y+=l->diry;
      insert_ob_in_map(l);
    }
  }
  hit_map(l,0,0);
}

void move_fb_bullet(object *b) {
  object *op;
  remove_ob(b);
  if(b->above!=NULL||blocked(b->x+b->dirx,b->y+b->diry)) {
    op=get_ball(b->stats.wc);
    op->owner=b->owner;
    op->x=b->x,op->y=b->y;
    if(!wall(b->x+b->dirx,b->y+b->diry))
      op->x+=b->dirx,op->y+=b->diry;
    free_object(b);
    insert_ob_in_map(op);
    return;
  }
  b->x+=b->dirx,b->y+=b->diry;
  insert_ob_in_map(b);
}

object *get_ball(int hp) {
  object *op=get_object();
  op->type=FBALL;
  op->face=G_BURN1;
  op->stats.hp=hp;
  op->stats.wc=2;
  op->stats.dam=2;
  op->speed=0.2,op->speed_left=-0.21;
  op->state=0;
  strncpy(op->name,"fireball",MAX_NAME);
  return op;
}

void fireball(object *b) {
  object *tmp;
  int i;

  remove_ob(b);
  if(--(b->stats.hp)<0) {
    free_object(b);
    return;
  }
  if(b->face==G_BURN1)
    b->face=G_BURN2;
  else
    b->face=G_BURN1;
  if(b->stats.hp>2&&!b->state) {
    b->state=1;
    for(i=1;i<9;i++) {
      int dx,dy;
      dx=b->x+freearr_x[i],dy=b->y+freearr_y[i];
      if(dx<0||dx>mapx||dy<0||dy>mapy||omap[dx][dy]!=G_BLANK)
        continue;
      if(present(FBALL,dx,dy)==NULL) {
        tmp=get_ball(b->stats.hp-1);
        tmp->owner=b->owner;
        tmp->x=dx,tmp->y=dy;
        insert_ob_in_map(tmp);
      }
    }
  }
  hit_map(b,0,0);
  insert_ob_in_map(b);
}

void hit_player(object *op,int dam, object *hitter) {
  char buf[MAX_BUF];

  if(op->wizard||op->type==BULLET||!op->alive||op->stats.hp<0) return;
  op->stats.hp-=dam;
  if(op->stats.hp<0) {
    if(op->type==DOOR) {
      op->speed=0.2;
      op->speed_left=0;
      return;
    }
    if(hitter->owner) {
      (void) sprintf(buf,"%s killed %s with %s.",hitter->owner->name,op->name,
              hitter->name);
      hitter=hitter->owner;
    }
    else
      (void) sprintf(buf,"%s killed %s.",hitter->name,op->name);
    add_exp(hitter,op->stats.exp/2);
/*
    info_all(buf);
*/
    if(hitter->type==PLAYER)
      draw_stats(hitter);
    if(op->type!=PLAYER) {
      remove_ob(op);
      free_object(op);
    }
  }
}

int blocked(int x,int y) {
  object *tmp;
  if(x<0||x>mapx||y<0||y>mapy) return 1;
  if(omap[x][y]!=G_BLANK) return 1;
  tmp=map_ob[x][y];
  while(tmp!=NULL) {
    if(tmp->alive) return 1;
    tmp=tmp->above;
  }
  return 0;
}

int wall(int x,int y) {
  if(x<0||x>mapx||y<0||y>mapy) return 1;
  if(omap[x][y]!=G_BLANK) return 1;
  return 0;
}

int inmap(int x,int y) {
  return !(x<0||x>mapx||y<0||y>mapy);
}

void move_player(object *op,int f,int dir,int dirm,char face) {
  object *tmp,*tmp2;

  if(f)
    fire(op,dir);
  else
    if(!move_ob(op,freearr_x[dir],freearr_y[dir])) {
      tmp=map_ob[op->x+freearr_x[dir]][op->y+freearr_y[dir]];
      while(tmp!=NULL&&!tmp->alive)
        tmp=tmp->above;
      if(tmp!=NULL) {
        if(tmp->type==DOOR) {
          tmp2=op->inv;
          while(tmp2!=NULL&&tmp2->type!=KEY)
            tmp2=tmp2->below;
          if(tmp2!=NULL) {
            remove_ob(tmp2);
            free_object(tmp2);
            hit_player(tmp,9999,op);
          }
        }
        if(tmp!=NULL)
          attack_ob(tmp,op);
      }
    }
  op->dirx=freearr_x[dirm],op->diry=freearr_y[dirm];
  op->face=face;
  update_object(op);
}

void pick_up(object *op) {
  object *tmp;
  char buf[MAX_BUF];

  if(op->below==NULL) {
    draw_info(op,"There is nothing to pick up here.");
    return;
  }
  tmp=op->below;
  (void) sprintf(buf,"You pick up the %s.",query_name(tmp));
  draw_info(op,buf);
  remove_ob(tmp); /* Unlink it */
  insert_ob_in_ob(tmp,op);
}

void drop(object *op) {
  object *tmp;
  char buf[MAX_BUF];

  tmp=op->inv;
  if(tmp==NULL) {
    draw_info(op,"You don't have anything to drop.");
    return;
  }
  op->inv=tmp->below;
  if(tmp->below!=NULL)
    tmp->below->above=NULL;
  remove_ob(tmp);
  tmp->x=op->x,tmp->y=op->y;
  insert_ob_in_map(tmp);
  (void) sprintf(buf,"You drop %s.",query_name(tmp));
  draw_info(op,buf);
  remove_ob(op);
  insert_ob_in_map(op);
}

void rotate(object *op) {
  object *tmp;
  char buf[MAX_BUF];

  tmp=op->inv;
  if(tmp==NULL) {
    draw_info(op,"You don't carry anything.");
    return;
  }
  while(tmp!=NULL&&tmp->below!=NULL)
    tmp=tmp->below;
  remove_ob(tmp);
  insert_ob_in_ob(tmp,op);
  sprintf(buf,"The %s (%d) is now on top.",query_name(tmp),tmp->count);
  draw_info(op,buf);
}

void apply(object *op) {
  object *tmp;
  char buf[MAX_BUF];

  tmp=op->inv;
  if(tmp==NULL) {
    draw_info(op,"You have nothing to apply.");
    return;
  }
  switch(tmp->type) {
  case POTION:
    draw_info(op,"You feel strange.");
    switch(random()%7) {
    case 0:
      if(op->stats.Con<25)
        op->stats.Con++;
      break;
    case 1:
      if(op->stats.Str<25)
        op->stats.Str++;
      break;
    case 2:
      if(op->stats.Dex<25)
        op->stats.Dex++;
      break;
    case 3:
      if(op->stats.Int<25)
        op->stats.Int++;
      break;
    case 4:
      if(op->stats.Cha<25)
        op->stats.Cha++;
      break;
    case 5:
      if(op->stats.Wis<25)
        op->stats.Wis++;
      break;
    case 6:
      switch(random()%6) {
      case 0:
        if(op->stats.Con>1)
          op->stats.Con--;
        break;
      case 1:
        if(op->stats.Str>1)
          op->stats.Str--;
        break;
      case 2:
        if(op->stats.Dex>1)
          op->stats.Dex--;
        break;
      case 3:
        if(op->stats.Int>1)
          op->stats.Int--;
        break;
      case 4:
        if(op->stats.Wis>1)
          op->stats.Wis--;
        break;
      case 5:
        if(op->stats.Cha>1)
          op->stats.Cha--;
        break;
      }
    }
    fix_player(op);
    remove_ob(tmp);
    free_object(tmp);
    return;
  case TREASURE:
    remove_ob(tmp);
    free_object(tmp);
    switch(random()%9) {
    case 0:
      sprintf(buf,"You find %d coins in the chest.",tmp->value);
      draw_info(op,buf);
      op->value+=tmp->value;
      draw_stats(op);
      return;
    case 1:
      draw_info(op,"You find a bow in the chest.");
      tmp=get_object();
      tmp->type=BOW;
      tmp->value=12;
      tmp->face=G_BOW;
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=12;
      strncpy(tmp->name,"bow",MAX_NAME);
      insert_ob_in_ob(tmp,op);
      return;
    case 2:
      draw_info(op,"You find some arrows in the chest.");
      tmp=get_object();
      tmp->type=ARROW;
      tmp->face=G_ARROW_1;
      tmp->nrof=2+random()%10;
      strncpy(tmp->name,"arrow",MAX_NAME);
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=1;
      insert_ob_in_ob(tmp,op);
      return;
    case 3:
      draw_info(op,"You find an axe in the chest.");
      tmp=get_object();
      tmp->type=WEAPON;
      tmp->face=G_AXE;
      tmp->stats.wc=(random()%3==0)?random()%4:0;
      sprintf(tmp->name,"axe +%d",tmp->stats.wc);
      tmp->stats.dam=4+tmp->stats.wc;
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=10;
      insert_ob_in_ob(tmp,op);
      return;
    case 4:
      draw_info(op,"You find a club in the chest.");
      tmp=get_object();
      tmp->type=WEAPON;
      tmp->face=G_CLUB;
      tmp->stats.wc=(random()%3==0)?random()%2:0;
      sprintf(tmp->name,"club +%d",tmp->stats.wc);
      tmp->stats.dam=2+tmp->stats.wc;
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=15;
      insert_ob_in_ob(tmp,op);
      return;
    case 5:
      draw_info(op,"You find a sword in the chest.");
      tmp=get_object();
      tmp->type=WEAPON;
      tmp->face=G_SWORD;
      tmp->stats.wc=(random()%3==0)?random()%5:0;
      sprintf(tmp->name,"sword +%d",tmp->stats.wc);
      tmp->stats.dam=6+tmp->stats.wc;
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=20;
      insert_ob_in_ob(tmp,op);
      return;
    case 6:
      draw_info(op,"You find a mace in the chest.");
      tmp=get_object();
      tmp->type=WEAPON;
      tmp->face=G_MACE;
      tmp->stats.wc=(random()%3==0)?random()%4:0;
      sprintf(tmp->name,"mace +%d",tmp->stats.wc);
      tmp->stats.dam=5+tmp->stats.wc;
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=20;
      insert_ob_in_ob(tmp,op);
      return;
    case 7:
      draw_info(op,"You find a ring mail in the chest.");
      tmp=get_object();
      tmp->type=ARMOUR;
      tmp->face=G_ARMOUR;
      tmp->stats.ac=random()%4+3;
      sprintf(tmp->name,"ring mail +%d",tmp->stats.ac);
      tmp->ox=op->x,tmp->oy=op->y;
      tmp->weight=tmp->stats.ac*10;
      insert_ob_in_ob(tmp,op);
      return;
    default:
      draw_info(op,"The chest was empty.");
      return;
    }
  case BOW:
    op->contr->shoottype=4;
    draw_info(op,"You will now fire arrows.");
    break;
  case WEAPON:
    if(op->weapon!=NULL) {
      draw_info(op,"You unwield your previous weapon.");
      insert_ob_in_ob(op->weapon,op);
    }
    op->weapon=tmp;
    remove_ob(tmp);
    sprintf(buf,"You now wield %s.",tmp->name);
    draw_info(op,buf);
    fix_player(op);
    break;
  case ARMOUR:
    if(op->armour!=NULL) {
      draw_info(op,"You take of your previous armour.");
      op->weight-=op->armour->weight;
      insert_ob_in_ob(op->armour,op);
    }
    op->armour=tmp;
    remove_ob(tmp);
    op->weight+=op->armour->weight;
    sprintf(buf,"You now wear %s.",tmp->name);
    draw_info(op,buf);
    fix_player(op);
    break;
  case FOOD:
    if(op->stats.food+tmp->stats.food>999) {
      draw_info(op,"You can't possible eat all that now.");
      break;
    }
    draw_info(op,"The food tasted good.");
    op->stats.food+=tmp->stats.food;
    op->stats.hp+=tmp->stats.food/50;
    if(op->stats.hp>op->stats.maxhp)
      op->stats.hp=op->stats.maxhp;
    draw_stats(op);
    remove_ob(tmp);
    free_object(tmp);
    break;
  case POISON:
    draw_info(op,"Yech!  That tasted poisonous!");
    op->stats.hp+=tmp->stats.hp;
    op->stats.food-=op->stats.food/4;
    draw_stats(op);
    remove_ob(tmp);
    free_object(tmp);
    break;
  default:
    sprintf(buf,"I don't know how to apply the %s (yet).",query_name(tmp));
    draw_info(op,buf);
  }
}

void inventory(object *op) {
  object *tmp;
  char buf[MAX_BUF];

  draw_info(op,"========================");
  if(op->weapon!=NULL) {
    sprintf(buf,"You wield %s. (+%d/+%d)",op->weapon->name,
            op->weapon->stats.wc, op->weapon->stats.dam);
    draw_info(op,buf);
  } else draw_info(op,"You wield nothing.");
  if(op->armour!=NULL) {
    sprintf(buf,"You wear %s. (%d)",op->armour->name,op->armour->stats.ac);
    draw_info(op,buf);
  } else draw_info(op,"You wear nothing.");
  if(op->inv==NULL)
    draw_info(op,"You carry nothing.");
  else {
    draw_info(op,"Inventory:");
    for(tmp=op->inv;tmp!=NULL;tmp=tmp->below) {
      (void) sprintf(buf,"- %s (%d).",query_name(tmp),tmp->count);
      draw_info(op,buf);
    }
  }
  sprintf(buf,"Total weight carried: %d.",op->weight);
  draw_info(op,buf);
}

void look(object *op) {
  object *tmp;
  char buf[MAX_BUF];
  int flag=0;

  if(op->above!=NULL) {
    flag=1;
    draw_info(op,"Above you see:");
    tmp=op->above;
    while(tmp!=NULL) {
      (void) sprintf(buf,"- %s (%d).",query_name(tmp),tmp->count);
      draw_info(op,buf);
      tmp=tmp->above;
    }
  }
  if(op->below!=NULL) {
    flag=1;
    draw_info(op,"Below you see:");
    tmp=op->below;
    while(tmp!=NULL) {
      (void) sprintf(buf,"- %s (%d).",query_name(tmp),tmp->count);
      draw_info(op,buf);
      tmp=tmp->below;
    }
  }
  if(!flag)
    draw_info(op,"You see nothing at all here.");
}

int parse_key(object *op,char k) {
  char buf[MAX_BUF];
  object *tmp;

  if(op->contr->writing) {
    if(k!=13) {
      write_ch(op,k);
      return 1;
    }
    if(op->contr->write_buf[0]=='"') {
      write_ch(op,'"');
      op->contr->writing=0;
      (void) sprintf(buf,"%s says %s",op->name,op->contr->write_buf);
      info_all(buf);
      info_flush();
      return 0;
    }
    if(op->contr->write_buf[0]=='>') {
      char *p;
      op->contr->writing=0;
      draw_info(op,op->contr->write_buf);
      p=strchr(op->contr->write_buf,' ');
      if(!strncmp(op->contr->write_buf,">add",4)) {
        if(p==NULL) {
          op->contr->writing=0;
          draw_info(op,"Add which display?");
          return 0;
        }
        if(add_player(++p))
          draw_info(op,errmsg);
        else
          draw_info(op,"OK.");
        return 0;
      }
      if(!strncmp(op->contr->write_buf,">name",5)) {
        if(p==NULL) {
          op->contr->writing=0;
          draw_info(op,"Change name to what?");
          return 0;
        }
        (void) sprintf(buf,"%s changes name to %s.",op->name,++p);
        strncpy(op->name,p,15);
        info_all(buf);
        op->contr->last_value=-1;
        draw_stats(op);
        return 0;
      }
      if(!strncmp(op->contr->write_buf,">who",4)) {
        int i;
        draw_info(op,"Players:");
        for(i=0;i<MAXPLAYERS;i++)
          if(players[i].used) {
            (void) sprintf(buf,"%s (%s)",players[i].ob->name,players[i].name);
            draw_info(op,buf);
          }
        return 0;
      }
      if(!strncmp(op->contr->write_buf,">cast",5)) {
        int i;
        if(p==NULL||!sscanf(++p,"%d",&i)) {
          op->contr->writing=0;
          draw_info(op,"Cast what?");
          return 0;
        }
        op->contr->shoottype=i;
        draw_info(op,"OK.");
        return 0;
      }
      if(!strncmp(op->contr->write_buf,">strength",9)) {
        int i;
        if(p==NULL||!sscanf(++p,"%d",&i)||i<3||i>op->stats.maxsp) {
          op->contr->writing=0;
          draw_info(op,"Set which strength?");
          return 0;
        }
        op->contr->shootstrength=i;
        draw_info(op,"OK.");
        return 0;
      }
      if(!strncmp(op->contr->write_buf,">dump",5)) {
        int i;
        if(p==NULL||!sscanf(++p,"%d",&i)||(tmp=find_object(i))==NULL) {
          op->contr->writing=0;
          draw_info(op,"Dump what object (nr)?");
          return 0;
        }
        dump_object(tmp);
        draw_info(op,errmsg);
        return 0;
      }
      if(!strncmp(op->contr->write_buf,">help",5)) {
        draw_info(op,"Commands:");
        draw_info(op,"  who           - Lists all players.");
        draw_info(op,"  add <display> - Adds a new player.");
        draw_info(op,"  name <name>   - Changes your name.");
        draw_info(op,"  cast <bulltyp>- Changes type of bul.");
        draw_info(op,"  strength <int>- Changes strength of bul.");
        draw_info(op,"  help          - Gives this info.");
        draw_info(op,"Push `hjklynub' to walk in a direction.");
        draw_info(op,"Shift + direction = fire, Ctrl + direction = run.");
        draw_info(op,"You damage monsters which you try to walk into.");
        draw_info(op,"`i' = inventory, `:' = look, `,' = pick, `d' = drop.");
        draw_info(op,"`>' = rotate inventory, `a' = apply top of inventory.");
        draw_info(op,"'+' = change spell.");
        draw_info(op,"Type `\"' to speak.");
        return 0;
      }
      draw_info(op,"Unknown command.  Try help.");
    }
    return 0;
  }
  switch(k) {
  case '"':
    write_ch(op,'"');
    return 0;
  case '\'':
    write_ch(op,'>');
    return 0;
  case ':': look(op); return 0;
  case 'Q': 
    remove_ob(op);
    op->state=1;
    sprintf(buf,"%s quits the game.",op->name);
    info_all(buf);
    info_flush();
    draw_info(op,"Do you want another game (y/n)?");
    return 0;
  case 'W':
    if(op->wizard) {
      op->wizard=0;
      draw_info(op,"Leaving Wizard mode.");
      return 0;
    }
    op->wizard=1;
    draw_info(op,"Entering Wizard mode.");
    return 0;
  case '>': rotate(op); return 0;
  case 'a': apply(op); return 0;
  case 'l': move_player(op,0,3,0,G_CHAR+1);return 0;
  case 'h': move_player(op,0,7,0,G_CHAR+0);return 0;
  case 'j': move_player(op,0,5,0,G_CHAR+3);return 0;
  case 'k': move_player(op,0,1,0,G_CHAR+2);return 0;
  case 'y': move_player(op,0,8,0,G_CHAR+2);return 0;
  case 'u': move_player(op,0,2,0,G_CHAR+1);return 0;
  case 'b': move_player(op,0,6,0,G_CHAR+0);return 0;
  case 'n': move_player(op,0,4,0,G_CHAR+3);return 0;
  case 'L': move_player(op,1,3,0,G_CHAR+1);return 0;
  case 'H': move_player(op,1,7,0,G_CHAR+0);return 0;
  case 'J': move_player(op,1,5,0,G_CHAR+3);return 0;
  case 'K': move_player(op,1,1,0,G_CHAR+2);return 0;
  case 'Y': move_player(op,1,8,0,G_CHAR+2);return 0;
  case 'U': move_player(op,1,2,0,G_CHAR+1);return 0;
  case 'B': move_player(op,1,6,0,G_CHAR+0);return 0;
  case 'N': move_player(op,1,4,0,G_CHAR+3);return 0;
  case 12 : move_player(op,0,3,3,G_CHAR+1);return 0;
  case  8 : move_player(op,0,7,7,G_CHAR+0);return 0;
  case 10 : move_player(op,0,5,5,G_CHAR+3);return 0;
  case 11 : move_player(op,0,1,1,G_CHAR+2);return 0;
  case 25 : move_player(op,0,8,8,G_CHAR+2);return 0;
  case 21 : move_player(op,0,2,2,G_CHAR+1);return 0;
  case 2  : move_player(op,0,6,6,G_CHAR+0);return 0;
  case 14 : move_player(op,0,4,4,G_CHAR+3);return 0;
  case '+':
    op->contr->shoottype++;
    if(op->contr->shoottype>3)
      op->contr->shoottype=0;
    switch(op->contr->shoottype) {
    case 0:
      draw_info(op,"Switched to bullets.");
      break;
    case 1:
      draw_info(op,"Switched to fireball.");
      break;
    case 2:
      draw_info(op,"Switched to lightning.");
      break;
    case 3:
      draw_info(op,"Switched to magic missile.");
      break;
    }
    return 0;
  case 'd': drop(op); return 0;
  case ',': pick_up(op); return 0;
  case 'i': inventory(op); return 0;
  case 'm': if (op->wizard) dump_map();  return 0;
  case 'M': if (op->wizard) dump_omap(); return 0;
  case 'P': if (op->wizard) dump_all_objects(); return 0;
  case 's': if (op->wizard)
    (void) fprintf(stderr,"%d used, %d free.\n",count_used(),count_free());
    return 0;
  }
  return 0;
}

void process_events() {
  int flag,repeat,i;
  object *op,*tmp;
  char text[10];
  char buf[MAX_BUF];

  for(flag=1;flag!=0;) {
    flag=0;
    for(op=objects;op!=NULL;op=op->next) {
      if(op->speed_left>0) {
        op->speed_left--;
        if(op->speed_left>0) flag=1; /* Another loop needed */
        if(op->alive&&(op->type==PLAYER&&!op->state)) {
          if(op->stats.Con&&--op->last_heal<op->stats.Con) {
            if(op->stats.hp<op->stats.maxhp)
              op->stats.hp+=op->stats.level,op->stats.food-=2;
            op->last_heal=27,
            op->stats.food--;
          }
          if(op->stats.Int&&--op->last_sp<op->stats.Int) {
            if(op->stats.sp<op->stats.maxsp)
              op->stats.sp+=op->stats.level,op->stats.food-=2;
            op->last_sp=27;
          }
        }
        switch(op->type) {
        case PLAYER:
          if (!op->state&&!op->wizard&&(op->stats.hp<0||op->stats.food<0)) {
            if(op->stats.hp<0)
              sprintf(buf,"%s died.",op->name);
            else
              sprintf(buf,"%s starved to death.",op->name);
            info_all(buf);
            info_flush();
            remove_ob(op);
            op->state=1;
            draw_info(op,"Do you want another game (y/n)?");
            break;
          }
          draw_stats(op);
          repeat=1;
          while(repeat) {
            repeat=0;
            if (XEventsQueued(op->contr->gdisp, QueuedAfterReading)==0) {
              if(op->dirx||op->diry)
                if(!move_ob(op,op->dirx,op->diry))
                  op->dirx=0,op->diry=0;
              XFlush(op->contr->gdisp);
              break;
            }
            XNextEvent(op->contr->gdisp,&op->contr->gevent);
            switch(op->contr->gevent.type) {
            case Expose:
              XClearWindow(op->contr->gdisp,op->contr->win_root);
              for(i=0;i<=WINLOWER-WINUPPER;i++)
                op->contr->drawn[0][i]=';';
              draw(op,op->x,op->y);
              op->contr->last_value=-1;
              draw_stats(op);
              draw_all_info(op);
              op->speed_left++;
              break;
            case MappingNotify:
              op->speed_left++;
              XRefreshKeyboardMapping(&op->contr->gevent);
            break;
            case KeyPress:
              if(XLookupString(&op->contr->gevent,text,10,
                               &op->contr->gkey,0)!=1)
                break;
              switch(op->state) {
              case 0:
                repeat=parse_key(op,text[0]);
                break;
              case 1:
                repeat=0;
                if(text[0]=='n') {
                  leave(op->contr->nr);
                  free_object(op);
                }
                else if(text[0]=='y') {
                  op->contr->ob=get_player(op->contr);
                  tmp=op->contr->ob;
                  tmp->state=2;
                  draw_info(tmp,"Welcome back!");
                  draw_info(tmp,"Roll again (y/n)?");
                  strncpy(tmp->name,op->name,MAX_NAME);
                  free_object(op);
                  op=tmp;
                }
                break;
              case 2:
                repeat=0;
                if(text[0]=='n') {
                  int j;
                  op->state=0;
                  j=find_free_spot(startx,starty,0,SIZEOFFREE);
                  op->x=startx+freearr_x[j],op->y=starty+freearr_y[j];
                  start_info(op);
                  insert_ob_in_map(op);
                  break;
                }
                roll_stats(op);
                draw_info(op,"Roll again (y/n)?");
                break;
              }
              continue;
            }
          }
          while(XEventsQueued(op->contr->gdisp, QueuedAfterReading))
            XNextEvent(op->contr->gdisp,&op->contr->gevent);
          break;
        case SCORP:
        case MUNCH:
        case BIRD:
        case GOB:
          move_monster(op);
          break;
        case GENGOB:
          generate_gob(op);
          break;
        case BULLET:
          move_bullet(op);
          break;
        case MMISSILE:
          move_missile(op);
          break;
        case ARROW:
          move_arrow(op);
          break;
        case FBULLET:
          move_fb_bullet(op);
          break;
        case FBALL:
          fireball(op);
          break;
        case LIGHTNING:
          move_lightning(op);
          break;
        case NEST:
          generate_bird(op);
          break;
        case GENSCORP:
          generate_scorp(op);
          break;
        case GENMUNCH:
          generate_munch(op);
          break;
        case DOOR:
          remove_door(op);
        }
      }
    }
  }
  for(i=0;i<MAXPLAYERS;i++)
    if(players[i].used) {
      draw(players[i].ob,players[i].ob->x,players[i].ob->y);
      XFlush(players[i].gdisp);
    }
  for(op=objects;op!=NULL;op=op->next)
    if(op->speed&&op->speed_left<=0)
      op->speed_left+=op->speed;
}

void leave(int i) {
  int j,k;
  char buf[MAX_BUF];

  XFlush(players[i].gdisp);
  XFreeGC(players[i].gdisp,players[i].gc_game);
  XDestroyWindow(players[i].gdisp,players[i].win_game);
  XCloseDisplay(players[i].gdisp);
  players[i].used=0;
/*
  remove_ob(players[i].ob);
  free_object(players[i].ob);
*/
  for(j=0,k=0;j<MAXPLAYERS;j++)
    if(players[j].used) k=1;
  if(!k) exit(0);
  (void) sprintf(buf,"%s left the game.",players[i].name);
  info_all(buf);
  info_flush();
}

int main(int argc,char **argv) {
  struct timeb tmp_time;
  unsigned short tmp_milli;

  fprintf(stderr,
    "Welcome to CrossFire, v%d.%d, copyright (C) 1992 Frank Tore Johansen.\n",
    VERSION,PATCH);
  gargc=argc, gargv=argv;
  init();
  while(1) {
    ftime(&tmp_time);
    tmp_milli=tmp_time.millitm;
    process_events();
    ftime(&tmp_time);
    tmp_milli+=262144-tmp_time.millitm;
    if(tmp_milli>=0)
      usleep(tmp_milli);
  }
  return 0;
}
