/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include "define.h"
#include "gauntlet.h"

object objarray[STARTMAX]; /* All objects, allocated this way at first */
object *objects;      /* Pointer to the list of used objects */
object *free_objects; /* Pointer to the list of unused objects */
/*
extern char map[MAX_X][MAX_Y];
extern char omap[MAX_X][MAX_Y];
extern int mapx,mapy;
extern object map_ob[MAX_X][MAX_Y];
*/

void dump_object(object *op) {
  if(op==NULL) {
    sprintf(errmsg,"[NULL pointer]");
    return;
  }
  sprintf(errmsg,"%d=%s [%2dx:%2d/%2d,y:%2d/%2d] sp:%1.2f/%1.2f",
    op->count,op->name,op->type,op->x,op->dirx,op->y,op->diry,
    op->speed_left,op->speed);
  if(op->is_freed)
    sprintf(errmsg,"%s (freed).",errmsg);
  else if(op->alive)
     sprintf(errmsg,"%s hp:%2d.",errmsg,op->stats.hp);
  else
     sprintf(errmsg,"%s (dead).",errmsg);
}

void dump_all_objects() {
  object *op;
  for(op=objects;op!=NULL;op=op->next) {
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
  }
}

object *find_object(int i) {
  object *op;
  for(op=objects;op!=NULL;op=op->next)
    if(op->count==i) return op;
  return NULL;
}

object *get_object() {
  object *op;
  if(free_objects==NULL) {
    fprintf(stderr,"No more objects left.\n");
    exit(1);
  }
  op=free_objects;
  if(!op->is_freed) {
    fprintf(stderr,"Fatal: Getting busy object.\n");
  }
  op->dirx=0,op->diry=0;
  free_objects=op->next;
  if(free_objects!=NULL)
    free_objects->prev=NULL;
  op->speed=0;
  op->speed_left=-2;
  op->next=objects;
  op->weapon=NULL;
  op->armour=NULL;
  op->prev=NULL;
  op->below=NULL;
  op->above=NULL;
  op->owner=NULL;
  op->inv=NULL;
  op->env=NULL;
  op->alive=0;
  op->weight=0;
  op->size=0;
  op->state=0;
  op->wizard=0;
  op->stats.level=0;
  op->value=0;
  op->last_heal=0;
  op->last_sp=0;
  op->is_removed=1;
  op->is_freed=0;
  op->count=++ob_count;
  op->stats.wc=0,op->stats.ac=0;
  op->stats.exp=0;
  op->stats.sp=0;
  op->stats.maxsp=0;
  op->stats.hp=0;
  op->stats.maxhp=0;
  op->stats.dam=0;
  op->stats.level=0;
  op->nrof=0;
  op->ox=0,op->oy=0;
  sprintf(op->name,"ob#%d",ob_count);
  if(objects!=NULL)
    objects->prev=op;
  objects=op;
  return op;
}

void update_object(object *op) {
  if(op->above==NULL)
    map[op->x][op->y]=op->face;
}

void free_object(object *ob) {
  object *tmp,*op;
  if(objects==NULL) {
    fprintf(stderr,"No object to be freed.\n");
    exit(1);
  }
  if(ob==NULL) {
    fprintf(stderr,"Trying to free NULL.\n");
    return;
  }
  /* First free the object from the used objects list: */
  if(ob->is_freed) {
    fprintf(stderr,"Trying to free freed object.\n");
    dump_object(ob);
    fprintf(stderr,"%s\n",errmsg);
    return;
  }
  if(!wall(ob->x,ob->y)) {
    if(ob->inv) {
      op=ob->inv;
      while(op!=NULL) {
        tmp=op->below;
        remove_ob(op);
        op->x=ob->x,op->y=ob->y;
        insert_ob_in_map(op);
        op=tmp;
      }
    }
    if(ob->weapon!=NULL) {
      ob->weapon->x=ob->x,ob->weapon->y=ob->y;
      insert_ob_in_map(ob->weapon);
    }
    if(ob->armour!=NULL) {
      ob->armour->x=ob->x,ob->armour->y=ob->y;
      insert_ob_in_map(ob->armour);
    }
  }
  ob->is_freed=1;
  if(ob->prev==NULL) {
    objects=ob->next;
    if(objects!=NULL)
      objects->prev=NULL;
  }
  else {
    ob->prev->next=ob->next;
    if(ob->next!=NULL)
      ob->next->prev=ob->prev;
  }
  
  /* Now link it with the free_objects list: */
  ob->prev=NULL;
  ob->next=free_objects;
  if(free_objects!=NULL)
    free_objects->prev=ob;
  free_objects=ob;
  ob->speed=0,ob->speed_left=-2; /* Just to be on the safe side... */
}

int count_free() {
  int i=0;
  object *tmp=free_objects;
  while(tmp!=NULL)
    tmp=tmp->next, i++;
  return i;
}

int count_used() {
  int i=0;
  object *tmp=objects;
  while(tmp!=NULL)
    tmp=tmp->next, i++;
  return i;
}

/* remove_ob(op):
 *   This function removes the object op from the linked list of objects
 *   which it is currently tied to.  When this function is done, the
 *   object will have no environment.  If the object previously had an
 *   environment, the x and y coordinates will be updated to
 *   the previous environment.
 */

void remove_ob(object *op) {
  if(op->is_removed) {
    fprintf(stderr,"Trying to remove removed object.\n");
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
    return;
  }
  if(op->x!=op->ox||op->y!=op->oy) {
    fprintf(stderr,"Object moved from %d,%d to %d,%d while not removed:\n",
            op->x,op->y,op->ox,op->oy);
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
    return;
  }
  if(op->x<0||op->x>mapx||op->y<0||op->y>mapy) {
    fprintf(stderr,"Trying to remove object outside the map.\n");
    return;
  }
  op->is_removed=1;
  if(op->env!=NULL) {
    if(op->nrof)
      op->env->weight-=op->weight*op->nrof;
    else
      op->env->weight-=op->weight;
    if(op->env->type==PLAYER)
      fix_player(op->env);
    if(op->above!=NULL)
      op->above->below=op->below;
    else
      op->env->inv=op->below;
    if(op->below!=NULL)
      op->below->above=op->above;
    op->x=op->env->x,op->y=op->env->y;
    op->ox=op->x,op->oy=op->y;
    op->env=NULL;
    op->above=NULL,op->below=NULL;
    return;
  }
  if(op->above!=NULL) {                /* Don't change map, we're not on top */
    op->above->below=op->below;        /* Link above with below */
    if(op->below!=NULL) {              /* Something below us? */
      op->below->above=op->above;      /* Yes, link above with below */
      op->below=NULL;
    }
    else {                             /* Nothing below, tell the floor what */
      if(map_ob[op->x][op->y]!=op) {
        fprintf(stderr,"Object squizez another object:\n");
        dump_object(op);
        fprintf(stderr,"%s\n",errmsg);
        dump_object(map_ob[op->x][op->y]);
        fprintf(stderr,"%s\n",errmsg);
      }
      map_ob[op->x][op->y]=op->above;  /* goes on above it. */
    }
    op->above=NULL;                      
  }
  else if(op->below!=NULL)
    map[op->x][op->y]=op->below->face, /* Show what we have below */
    op->below->above=NULL,             /* It is now on top */
    op->below=NULL;
  else {
      if(map_ob[op->x][op->y]!=op) {
        fprintf(stderr,"Object squizez another object:\n");
        dump_object(op);
        fprintf(stderr,"%s\n",errmsg);
        dump_object(map_ob[op->x][op->y]);
        fprintf(stderr,"%s\n",errmsg);
      }
    map[op->x][op->y]=omap[op->x][op->y],/* Show the floor */
    map_ob[op->x][op->y]=NULL;           /* No more objects here */
  }
}

/* insert_ob_in_ob(op,environment):
 *   This function inserts the object op in the linked list
 *   inside the object environment.  If environment is NULL,
 *   the object will be stacked at the map.
 */

void insert_ob_in_map(object *op) {
  object *tmp;
  if(op->x<0||op->x>mapx||op->y<0||op->y>mapy) {
    fprintf(stderr,"Trying to insert object outside the map.\n");
    return;
  }
  if(!op->is_removed) {
    fprintf(stderr,"Trying to insert (map) inserted object.\n");
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
    return;
  }
  if(op->nrof&&!op->speed)
    for(tmp=map_ob[op->x][op->y];tmp!=NULL;tmp=tmp->above)
      if(tmp->type==op->type&&!tmp->speed) {
        op->nrof+=tmp->nrof;
        remove_ob(tmp);
        free_object(tmp);
      }
  op->ox=op->x,op->oy=op->y;
  op->is_removed=0;
  map[op->x][op->y]=op->face;            /* Change the appearance of the map */
  if((tmp=map_ob[op->x][op->y])!=NULL) { /* If anything is below, then */
    while(tmp->above!=NULL)              /* Find out what is on top */
      tmp=tmp->above;
    op->below=tmp,                       /* link it. */
    tmp->above=op;                       /* (Or rather, put us on top of it) */
  }
  else
    map_ob[op->x][op->y]=op;             /* Tell the map that we're here */
}

object *get_split_ob(object *where,int type) {
  object *tmp,*op;
  for(tmp=where->inv;tmp!=NULL;tmp=tmp->below)
    if(tmp->type==type) {
      if(!tmp->nrof) {
        fprintf(stderr,"Trying to split non-splittable object.\n");
        return NULL;
      }
      if(!--tmp->nrof) {
        remove_ob(tmp);
        return tmp;
      }
      else
        where->weight--;
      op=get_object();
      op->type=type,op->face=tmp->face;
      op->value=tmp->value;
      op->ox=tmp->x,op->x=tmp->x;
      op->oy=tmp->y,op->y=tmp->y;
      return op;
    }
  return NULL;
}

int split_ob(object *where, int type) {
  object *tmp;
  for(tmp=where->inv;tmp!=NULL;tmp=tmp->below)
    if(tmp->type==type) {
      if(!tmp->nrof) {
        fprintf(stderr,"Trying to split non-splittable object.\n");
        return 0;
      }
      if(--(tmp->nrof)<1) {
        remove_ob(tmp);
        free_object(tmp);
      }
      else
        where->weight--;
      return 1;
    }
  return 0;
}

void insert_ob_in_ob(object *op,object *where) {
  object *tmp;
  if(!op->is_removed) {
    fprintf(stderr,"Trying to insert (ob) inserted object.\n");
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
    return;
  }
  if(where==NULL) {
    fprintf(stderr,"Trying to put object in NULL.\n");
    dump_object(op);
    fprintf(stderr,"%s\n",errmsg);
  }
  if(op->nrof) {
    for(tmp=where->inv;tmp!=NULL;tmp=tmp->below)
      if(tmp->type==op->type) {
        op->nrof+=tmp->nrof;
        remove_ob(tmp);
        free_object(tmp);
      }
    where->weight+=op->weight*op->nrof;
  } else
    where->weight+=op->weight;
  if(where->type==PLAYER)
    fix_player(where);
  op->is_removed=0;
  op->env=where;
  op->above=NULL;
  if(where->inv!=NULL)
    op->below=where->inv,
    op->below->above=op;
  else
    op->below=NULL;
  where->inv=op;
}

int move_ob(object *op,int dx,int dy) {
  if(!op->wizard&&blocked(op->x+dx,op->y+dy)) return 0;
  remove_ob(op);                        /* Removes it from the map */
  op->x+=dx,op->y+=dy;                  /* Move this object */
  insert_ob_in_map(op);
  return 1;
}

int hit_map(object *op,int dx,int dy) {
  object *tmp;

  if(op->x+dx<0||op->x+dx>mapx||op->y+dy<0||op->y+dy>mapy) return 0;
  if(!op->wizard&&omap[op->x+dx][op->y+dy]!=G_BLANK) return 0;
  for(tmp=map_ob[op->x+dx][op->y+dy];tmp!=NULL;tmp=tmp->above)
    if(tmp->alive)
      hit_player(tmp,op->stats.dam,op);
  if(tmp==NULL) return 0;
  return 1;
}

object *present(char type,int x,int y) {
  object *tmp;
  if(x<0||x>mapx||y<0||y>mapy) {
    fprintf(stderr,"Present called outside map.\n");
    return NULL;
  }
  for(tmp=map_ob[x][y];tmp!=NULL;tmp=tmp->above)
    if(tmp->type==type)
      return tmp;
  return NULL;
}

char *query_name(object *op) {
  if(op->nrof) {
    if(op->nrof>1)
      sprintf(errmsg,"%d %ss",op->nrof,op->name);
    else
      sprintf(errmsg,"%d %s",op->nrof,op->name);
    return errmsg;
  }
  return op->name;
}
