/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include <strings.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>

typedef struct liv {
  int Str,Dex,Con,Wis,Cha,Int;
  int hp;   /* Hit Points.  When 0 is reached, the object will be deleted. */
  int maxhp;
  int sp;   /* Spell points.  Used to cast spells. */
  int maxsp;/* Max spell points. */
  int exp;  /* Experience.  Killers gain 1/10. */
  int food; /* How much food in stomach.  When it reaches 0, you're dead. */
  int wc,ac;/* Weapon Class and Armour Class */
  int dam;  /* How much damage this object does when hitting */
  int levhp[11]; /* What the player gained on that level */
  int levsp[11];
  int level;
} living;


typedef struct obj {
  char name[16];
  int nrof;
  int x,y;             /* Position in the map for this object */
  int ox,oy;           /* For debugging: Where it was last inserted */
  float speed;         /* The overall speed of this object */
  float speed_left;    /* How much speed is left to spend this round */
  int count;           /* Which nr. of object created this is. */
  int dirx,diry;       /* Means the object is moving that way. */
  int type;            /* PLAYER, BULLET, etc */
  int weight, size;    /* Attributes of the object */
  int value;           /* How much money it is worth (or contains) */
  int last_heal;       /* Last healed. Depends on constitution */
  int last_sp;         /* As last_heal, but for spell points */
  int carrying;        /* How much weight this object contains */
  char face;
  struct obj *weapon;  /* Weapon wielded */
  struct obj *armour;  /* Armour worn */
  struct obj *next;    /* Pointer to the next object in the free/used list */
  struct obj *prev;    /* Pointer to the previous object in the free/used list*/
  struct obj *below;   /* Pointer to the object stacked below this one */
  struct obj *above;   /* Pointer to the object stacked above this one */
                       /* Note: stacked in the *same* environment! */
  struct obj *inv;     /* Pointer to the first object in the inventory */
  struct obj *env;     /* Pointer to the object which is the environment */
  struct obj *owner;   /* Pointer to the object which controls this one */
  struct pl *contr;    /* Pointer to the player which control this object */
  int alive;           /* All objects which want to fight must be alive */
  int state;           /* How the object was last drawn (animation) */
  living stats;        /* I'll make this a pointer when I've made the
                          allocation system */
/* Some debug variables: */
  int wizard;          /* If true, this object can break some rules */
  int is_removed;      /* If true, remove_ob() has been called */
  int is_freed;        /* If true, free_object() has been called */
} object;

void remove_ob();
object *get_split_ob();
int wall();
void fix_player();
