/*
    CrossFire, A Multiplayer game for X-windows

    Copyright (C) 1992 Frank Tore Johansen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author can be reached via e-mail to frankj@ifi.uio.no.
*/

#include "gauntlet.h"
#include "xio.h"

/*
extern object *get_object();
extern void free_object();
extern int count_free();
extern int count_used();
extern void remove_ob();
extern void insert_ob();
extern int move_ob();
*/

extern object *objects;
extern object *free_objects;
extern object objarray[STARTMAX];

static char name[]="Gauntlet";

int fixfontpath(player *pl) {
  int count;
  char **list;

  list=XListFonts(pl->gdisp,"gauntlet",1,&count);
  XFreeFontNames(list);
  if(count) return 0;
  sprintf(errmsg,"Display %s doesn't have the right fontpath.",pl->name);
  return 1;
}

int IOerrors(Display *d) {
  fprintf(stderr,"Fatal error on display.\n");
  return 0;
}

int Xerrors(Display *d,XErrorEvent *err) {
  char buf[MAX_BUF];
  XGetErrorText(d,err->error_code,buf,MAX_BUF);
  (void) fprintf(stderr,"Error code %s\n",buf);
  return 0;
}

int get_root_display(player *p, char *display) {
  p->gdisp=XOpenDisplay(display);
  if(!p->gdisp) {
    sprintf(errmsg,"Can't open display %s.",display);
    return 1;
  }
  p->gscreen=DefaultScreen(p->gdisp);
  p->gbackground=BlackPixel(p->gdisp,p->gscreen);
  p->gforeground=WhitePixel(p->gdisp,p->gscreen);
  p->roothint.x=0,p->roothint.y=0;
  p->roothint.width=570,p->roothint.height=385;
  p->win_root=XCreateSimpleWindow(p->gdisp,DefaultRootWindow(p->gdisp),
      p->roothint.x,p->roothint.y,p->roothint.width,p->roothint.height,2,
      p->gforeground,p->gbackground);
  XSetStandardProperties(p->gdisp,p->win_root,name,name,
      None,gargv,gargc,&(p->roothint));
  p->gc_root=XCreateGC(p->gdisp,p->win_root,0,0);
  XSetForeground(p->gdisp,p->gc_root,p->gbackground);
  XSetBackground(p->gdisp,p->gc_root,p->gforeground);
  if(fixfontpath(p))
    return 1;
  p->font=XLoadFont(p->gdisp,font_graphic);
  XSetFont(p->gdisp,p->gc_root,p->font);
  XSelectInput(p->gdisp,p->win_root,
    KeyPressMask|ExposureMask|StructureNotifyMask);
  XMapRaised(p->gdisp,p->win_root);
  XNextEvent(p->gdisp,&p->gevent);
  return 0;
}

int get_game_display(player *p,char *display) {
  p->gamehint.x=5,p->gamehint.y=110;
  p->gamehint.width=269,p->gamehint.height=269;
  p->gamehint.flags=PPosition | PSize;
  p->win_game=XCreateSimpleWindow(p->gdisp,p->win_root,
               p->gamehint.x,p->gamehint.y,
               p->gamehint.width,p->gamehint.height,2,
               p->gforeground,p->gbackground);
  XSetStandardProperties(p->gdisp,p->win_game,name,name,None,gargv,gargc,
                         &(p->gamehint));
  p->gc_game=XCreateGC(p->gdisp,p->win_game,0,0);
  XSetForeground(p->gdisp,p->gc_game,p->gbackground);
  XSetBackground(p->gdisp,p->gc_game,p->gforeground);
  p->font=XLoadFont(p->gdisp,font_graphic);
  XSetFont(p->gdisp,p->gc_game,p->font);
  XSelectInput(p->gdisp,p->win_game,KeyPressMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_game);
  return 0;
}

int get_info_display(player *p,char *name) {
  p->infohint.x=283,p->infohint.y=3;
  p->infohint.width=280,p->infohint.height=378;
  p->infohint.flags=PPosition | PSize;
  p->win_info=XCreateSimpleWindow(p->gdisp,p->win_root,
      p->infohint.x,p->infohint.y,p->infohint.width,p->infohint.height,2,
      p->gbackground,p->gforeground);
  XSetStandardProperties(p->gdisp,p->win_info,name,name,None,gargv,gargc,
                         &(p->infohint));
  p->gc_info=XCreateGC(p->gdisp,p->win_info,0,0);
  XSetForeground(p->gdisp,p->gc_info,p->gbackground);
  XSetBackground(p->gdisp,p->gc_info,p->gforeground);
  p->font=XLoadFont(p->gdisp,font_text_info);
  XSetFont(p->gdisp,p->gc_info,p->font);
  XSelectInput(p->gdisp,p->win_info,KeyPressMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_info);
  return 0;
}

int get_stats_display(player *p,char *name) {
  p->stathint.x=3,p->stathint.y=3;
  p->stathint.width=272,p->stathint.height=100;
  p->stathint.flags=PPosition | PSize;
  p->win_stats=XCreateSimpleWindow(p->gdisp,p->win_root,
      p->stathint.x,p->stathint.y,p->stathint.width,p->stathint.height,2,
      p->gbackground,p->gforeground);
  XSetStandardProperties(p->gdisp,p->win_stats,name,name,None,gargv,gargc,
                         &(p->stathint));
  p->gc_stats=XCreateGC(p->gdisp,p->win_stats,0,0);
  XSetForeground(p->gdisp,p->gc_stats,p->gbackground);
  XSetBackground(p->gdisp,p->gc_stats,p->gforeground);
  p->font=XLoadFont(p->gdisp,font_text_stats);
  XSetFont(p->gdisp,p->gc_stats,p->font);
  XSelectInput(p->gdisp,p->win_stats,KeyPressMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_stats);
  return 0;
}

void draw_stats(object *pl) {
  char buff[6][MAX_BUF];
  int flag[6];
  int i;
  for(i=0;i<6;i++)
    flag[i]=0;
  if(pl->contr->last_value==-1||
     pl->contr->last_stats.level!=pl->stats.level) {
    sprintf(buff[0],"Player: %s  Level: %d",pl->name,pl->stats.level);
    flag[0]=1;
    pl->contr->last_stats.level=pl->stats.level;
  }
  if(pl->contr->last_value==-1||pl->stats.exp!=pl->contr->last_stats.exp||
     pl->value!=pl->contr->last_value) {
    sprintf(buff[1],"Score: %4d  Coins: %4d",pl->stats.exp,pl->value);
    flag[1]=1;
    pl->contr->last_stats.exp=pl->stats.exp;
  }
  if(pl->contr->last_value==-1||
     pl->stats.hp!=pl->contr->last_stats.hp||
     pl->stats.maxhp!=pl->contr->last_stats.maxhp||
     pl->stats.sp!=pl->contr->last_stats.sp||
     pl->stats.maxsp!=pl->contr->last_stats.maxsp) {
    sprintf(buff[2],"Hp: %d/%d  Sp: %d/%d",
            pl->stats.hp,pl->stats.maxhp,pl->stats.sp,pl->stats.maxsp,
            pl->stats.food);
    flag[2]=1;
    pl->contr->last_stats.hp=pl->stats.hp;
    pl->contr->last_stats.maxhp=pl->stats.maxhp;
    pl->contr->last_stats.sp=pl->stats.sp;
    pl->contr->last_stats.maxsp=pl->stats.maxsp;
  }
  if(pl->contr->last_value==-1||
    pl->stats.Dex!=pl->contr->last_stats.Dex ||
    pl->stats.Con!=pl->contr->last_stats.Con ||
    pl->stats.Str!=pl->contr->last_stats.Str ||
    pl->stats.Int!=pl->contr->last_stats.Int ||
    pl->stats.Wis!=pl->contr->last_stats.Wis ||
    pl->stats.Cha!=pl->contr->last_stats.Cha) {
    sprintf(buff[3],"St%2d Dx%2d Co%2d In%2d Wi%2d Ch%2d",
            pl->stats.Str,pl->stats.Dex,pl->stats.Con,
            pl->stats.Int,pl->stats.Wis,pl->stats.Cha);
    flag[3]=1;
    pl->contr->last_stats.Str=pl->stats.Str;
    pl->contr->last_stats.Con=pl->stats.Con;
    pl->contr->last_stats.Dex=pl->stats.Dex;
    pl->contr->last_stats.Int=pl->stats.Int;
    pl->contr->last_stats.Wis=pl->stats.Wis;
    pl->contr->last_stats.Cha=pl->stats.Cha;
  }
  if(pl->contr->last_value==-1||
     pl->stats.wc!=pl->contr->last_stats.wc||
     pl->stats.ac!=pl->contr->last_stats.ac||
     pl->stats.dam!=pl->contr->last_stats.dam) {
     sprintf(buff[4],"Thaco: %2d  Damage: %2d  Ac: %2d",
             pl->stats.wc,pl->stats.dam,pl->stats.ac);
     flag[4]=1;
     pl->contr->last_stats.wc=pl->stats.wc;
     pl->contr->last_stats.ac=pl->stats.ac;
     pl->contr->last_stats.dam=pl->stats.dam;
  }
  if(pl->contr->last_value==-1||pl->contr->last_speed!=pl->speed||
     pl->stats.food!=pl->contr->last_stats.food) {
     flag[5]=1;
     if(pl->stats.food<100&&(pl->stats.food&4))
       sprintf(buff[5],"Speed: %3.1f  Food: *%d* HUNGRY!",
               pl->speed,pl->stats.food);
     else
       sprintf(buff[5],"Speed: %3.1f  Food: %3d",
               pl->speed,pl->stats.food);
     pl->contr->last_stats.food=pl->stats.food;
     pl->contr->last_speed=pl->speed;
  }
  for(i=0;i<6;i++)
    if(flag[i]) {
      strcat(buff[i],"                     "); /* In case it became shorter */
      XDrawImageString(pl->contr->gdisp,pl->contr->win_stats,
        pl->contr->gc_stats,10,i*14+10, buff[i],strlen(buff[i]));
    }
  pl->contr->last_value=pl->value;
}

void draw_all_info(object *pl) {
  int i;
  XClearWindow(pl->contr->gdisp,pl->contr->win_info);
  for(i=0;i<INFOLINES;i++)
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,
     pl->contr->gc_info,10,i*8+8,pl->contr->info[i],strlen(pl->contr->info[i]));
}

void info_all(char *str) {
  int i;
  char buf[MAX_BUF];

  if(ann_count) {
    if(!strncmp(str,ann_last,MAX_BUF)) {
      ann_count++;
      return;
    }
    if(ann_count>1)
      sprintf(buf,"%d times %s",ann_count,ann_last);
    else
      strncpy(buf,ann_last,MAX_BUF);
    for(i=0;i<MAXPLAYERS;i++)
      if(players[i].used)
        draw_info(players[i].ob,buf);
  }
  strncpy(ann_last,str,MAX_BUF);
  ann_count=1;
}

void info_flush() {
  info_all("blah");
  ann_count=0;
}

void draw_info(object *pl,char *str) {
  static char blanks[]="                                                      ";
  strncpy(pl->contr->info[pl->contr->infoline],str,INFOCHARS);
  XDrawImageString(pl->contr->gdisp,pl->contr->win_info,
    pl->contr->gc_info,10,pl->contr->infoline*8+8,
    pl->contr->info[pl->contr->infoline],
    strlen(pl->contr->info[pl->contr->infoline]));
  if(++(pl->contr->infoline)>=INFOLINES)
    pl->contr->infoline=0;
  strncpy(pl->contr->info[pl->contr->infoline]," ",INFOCHARS);
  XDrawImageString(pl->contr->gdisp,pl->contr->win_info,
    pl->contr->gc_info,10,pl->contr->infoline*8+8,blanks,strlen(blanks));
  if(pl->contr->writing) {
    strncpy(pl->contr->info[pl->contr->infoline],
            pl->contr->write_buf,INFOCHARS);
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
      10,pl->contr->infoline*8+8,pl->contr->write_buf,pl->contr->writing);
  }
}

void write_ch(object *pl,char key) {
  char *c;

  if((key<32||key>127)&&key!=8) return;
  c=&pl->contr->write_buf[pl->contr->writing];
  if(key==8||key==127) {
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
      10+(--pl->contr->writing)*5,pl->contr->infoline*8+8," ",1);
  }
  else
    pl->contr->write_buf[pl->contr->writing++]=key;
  pl->contr->write_buf[pl->contr->writing]='\0';
  if(key!=8&&key!=127)
    XDrawImageString(pl->contr->gdisp,pl->contr->win_info,pl->contr->gc_info,
      5+pl->contr->writing*5,pl->contr->infoline*8+8,c,1);
}

void draw(object *pl,int x,int y) {
  int i,j,left,right,ax,ay; /* ax and ay goes from 0 to max-size of arrays */
  char buff[20],tmp;
  buff[12]='\0';
  for(j=y+WINUPPER;j<=y+WINLOWER;j++) {
    left=-1,right=0,ay=j-y-WINUPPER;
    for(i=x+WINLEFT;i<=x+WINRIGHT;i++) {
      ax=i-x-WINLEFT;
      if((tmp=(buff[ax]=(i<0||i>mapx||j<0||j>mapy)?G_BLANK:map[i][j]))
         !=pl->contr->drawn[ax][ay]) {
          pl->contr->drawn[ax][ay]=tmp;
	  if(left==-1) left=ax;
          right=ax;
      }
    }
    if(left!=-1)
      XDrawImageString(pl->contr->gdisp,pl->contr->win_game,
        pl->contr->gc_game,2+24*left,ay*24+26, buff+left,right-left+1);
  }
}
